__all__: list[str] = []

# Enumerations
AccelType_HOST: int
ACCEL_TYPE_HOST: int
AccelType_DX11: int
ACCEL_TYPE_DX11: int
AccelType_VAAPI: int
ACCEL_TYPE_VAAPI: int
AccelType_LAST_VALUE: int
ACCEL_TYPE_LAST_VALUE: int
AccelType = int
"""One of [AccelType_HOST, ACCEL_TYPE_HOST, AccelType_DX11, ACCEL_TYPE_DX11, AccelType_VAAPI, ACCEL_TYPE_VAAPI, AccelType_LAST_VALUE, ACCEL_TYPE_LAST_VALUE]"""



