// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html
#ifndef OPENCV_SIGNAL_HPP
#define OPENCV_SIGNAL_HPP

/**
* @defgroup signal Signal Processing
* @{
* This module includes signal processing algorithms.
* @}
*/

#include "opencv2/core.hpp"
#include "opencv2/signal/signal_resample.hpp"

#endif
