__all__: list[str] = []

ParamDesc_Kind_Load: int
PARAM_DESC_KIND_LOAD: int
ParamDesc_Kind_Import: int
PARAM_DESC_KIND_IMPORT: int
ParamDesc_Kind = int
"""One of [ParamDesc_Kind_Load, PARAM_DESC_KIND_LOAD, ParamDesc_Kind_Import, PARAM_DESC_KIND_IMPORT]"""


# Classes

