__all__: list[str] = []

# Enumerations
MM_TRANSLATION: int
MM_TRANSLATION_AND_SCALE: int
MM_ROTATION: int
MM_RIGID: int
MM_SIMILARITY: int
MM_AFFINE: int
MM_HOMOGRAPHY: int
MM_UNKNOWN: int
MotionModel = int
"""One of [MM_TRANSLATION, MM_TRANSLATION_AND_SCALE, MM_ROTATION, MM_RIGID, MM_SIMILARITY, MM_AFFINE, MM_HOMOGRAPHY, MM_UNKNOWN]"""



